package ptkobjform ; 

require Exporter;   
@ISA = ('Exporter');

@EXPORT=qw(ptkform ptksticky) ;

use Tk ;
use Tk::LabEntry ;
use Tk::Balloon ;
use Tk::ROText ;

our (@Ggrid, @Gcol, @Grow, @Gframe, %Value, $Madefonts, $Sticky, $Verbose) ; 

$Sticky = "nw" ; 

sub mkfonts {
    my ($MW) = @_ ; 
    if ( ! $Madefonts ) { 

	$MW->fontCreate("ptkformLabel", 
	    -family=>"lucida",	     -size=>9) ;
	$MW->fontCreate("ptkformEmphasize", 
	    -family=>"lucida",	     -size=>16) ;
	$MW->fontCreate("ptkformEntry", 
	    -family=>'lucida typewriter',         -weight=>'bold', -size=>12) ;
	$MW->fontCreate("ptkformVerysmall", 
	    -family=>'palatino',         -size=>3) ;

	$Madefonts = 1 ;
    }
}

sub ptksticky { 
    my $orig = $Sticky ;
    $Sticky = $_[0] if defined $_[0] ; 
    return $Sticky ; 
}

sub parserc { 
    my ($rc) = @_ ; 
    my ($row, $col) = split(',', $rc ) ; 
    $G{row} = 0 if ! defined $G{row} ;
    $G{col} = 0 if ! defined $G{col} ;
    if ( $row =~ /(\S+):(\S+)/ ) { 
	$row = $1 ;
	$rs = $2 ;
    } else { 
	$rs = 1 ; 
    } 
    if ( $row eq "=" ) {
	;
    } elsif ( $row =~ /\+(\d+)?/ ) { 
	if ( defined $1 ) { 
	    $G{row} += $1 ; 
	} else { 
	    $G{row} ++ ;
	}
    } elsif ( $row =~ /\-(\d+)?/ ) { 
	if ( defined $1 ) { 
	    $G{row} -= $1 ; 
	} else { 
	    $G{row} -- ;
	}
    } elsif ( $row =~  /\d+/ ) { 
	$G{row} = $row ; 
    }
    $row = $G{row} ; 
    $G{row} += $rs-1 ;

    if ( $col =~ /(\S+):(\S+)/ ) { 
	$col = $1 ;
	$cs = $2 ;
    } else { 
	$cs = 1 ; 
    }
    if ( $col eq "=" ) {
	;
    } elsif ( $col =~ /\+(\d+)?/ ) { 
	if ( defined $1 ) { 
	    $G{col} += $1 ; 
	} else { 
	    $G{col} ++ ;
	}
    } elsif ( $col =~ /\-(\d+)?/ ) { 
	if ( defined $1 ) { 
	    $G{col} -= $1 ; 
	} else { 
	    $G{col} -- ;
	}
    } elsif ( $col =~  /\d+/ ) { 
	$G{col} = $col ; 
    }
    $col = $G{col} ; 
    $G{col} += $cs-1 ;
    return ($row, $rs, $col, $cs) ;
}

sub ptkEntry {
    my ( $f, $ref, $label, $name, $width, $widgets) = @_ ; 
    my $s = $f->Frame() ; 
    my $l = $s->Label(-text=>$label, -font=>"ptkformLabel", -fg=>"darkblue")  
	    ->pack(-side=>"top", -ipady=>0, -pady=>0, -anchor=>"sw" ) ; 
    $ref->$name = "" if ! defined $ref->$name ; 
    my $a = \$ref->$name ;
    my $e = $s->Entry(-textvariable=>$a, -font=>"ptkformEntry") 
	->pack(-side=>"top", -ipady=>0, -pady=>0, -anchor=>"nw" ) ; 
    if ( $width ne "-" ) { 
	$e->configure(-width=>$width) ; 
    }
    $e->bind('<Control-u>' => sub { $e->delete(0,"end") } ) ; 
    if ( defined $$widgets{$name} ) { 
	print STDERR "widget{$name} multiply defined.\n" 
    }
    $$widgets{"l_$name"} = $l ;
    $$widgets{$name} = $e ;
    my $spacer = $s->Label(-font=>"ptkformVerysmall")
	    ->pack(-side=>"top", -ipady=>0, -pady=>0, -anchor=>"nw" ) ; 
    return $s ; 
}

sub ptkScale {
    my ( $f, $ref, $label, $name, $width, $widgets, $orient) = @_ ; 
    my ($from, $to, $l ) = split (' ', $label, 3) ;
    my $digits=0 ; 
    my $x ;
    if ( $from < $to ) { 
	$x = $from ; 
    } else { 
	$x = $to ; 
    }
    while ( $x < 1 && $x != 0 ) { 
	$digits++ ; 
	$x *= 10 ; 
    }
    my $s = $f->Frame() ; 
    my $l = $s->Label(-text=>$l, -font=>"ptkformLabel", -fg=>"darkblue")  
	    ->pack(-side=>"top", -ipady=>0, -pady=>0, -anchor=>"sw" ) if $l !~ /^\s*$/ ; 
    $ref->$name = "" if ! defined $ref->$name ; 
    my $a = \$ref->$name ;
    my $e = $s->Scale(-variable=>$a, -font=>"ptkformEntry", -orient=>$orient,
			-from=>$from, -to=>$to, -digits=>$digits ) 
	->pack(-side=>"top", -ipady=>0, -pady=>0, -anchor=>"nw" ) ; 
    if ( $width ne "-" ) { 
	$e->configure(-length=>$width) ; 
    }
    if ( defined $$widgets{$name} ) { 
	print STDERR "widget{$name} multiply defined.\n" 
    }
    $$widgets{$name} = $e ;
    my $spacer = $s->Label(-font=>"ptkformVerysmall")
	    ->pack(-side=>"top", -ipady=>0, -pady=>0, -anchor=>"nw" ) ; 
    return $s ; 
}

sub ptkROText {
    my ( $f, $ref, $label, $name, $width, $widgets) = @_ ; 
    my $s = $f->Frame() ; 
    my $l = $s->Label(-text=>$label, -font=>"ptkformLabel", -fg=>"darkblue")  
	    ->pack(-side=>"top", -ipady=>0, -pady=>0, -anchor=>"sw" ) ; 
    my $e = $s->Scrolled("ROText", -scrollbars=>"osoe", -font=>"ptkformEntry") 
	->pack(-side=>"top", -ipady=>0, -pady=>0, -anchor=>"nw" ) ; 
    if ( $width ne "-" ) { 
	my ($rows, $columns) = split ('x', $width) ;
	$e->configure(-height=>$rows) ; 
	if ( defined $columns ) { 
	    $e->configure(-width=>$columns) ; 
	}
    }
    if ( defined $$widgets{$name} ) { 
	print STDERR "widget{$name} multiply defined.\n" 
    }
    $$widgets{$name} = $e ;
    my $spacer = $s->Label(-font=>"ptkformVerysmall")
	    ->pack(-side=>"top", -ipady=>0, -pady=>0, -anchor=>"nw" ) ; 
    return $s ; 
}

sub ptkText {
    my ( $f, $ref, $label, $name, $width, $widgets) = @_ ; 
    my $s = $f->Frame() ; 
    my $l = $s->Label(-text=>$label, -font=>"ptkformLabel", -fg=>"darkblue")  
	    ->pack(-side=>"top", -ipady=>0, -pady=>0, -anchor=>"sw" ) ; 
    my $e = $s->Scrolled("Text", -scrollbars=>"osoe", -font=>"ptkformEntry") 
	->pack(-side=>"top", -ipady=>0, -pady=>0, -anchor=>"nw" ) ; 
    if ( $width ne "-" ) { 
	my ($rows, $columns) = split ('x', $width) ;
	$e->configure(-height=>$rows) ; 
	if ( defined $columns ) { 
	    $e->configure(-width=>$columns) ; 
	}
    }
    if ( defined $$widgets{$name} ) { 
	print STDERR "widget{$name} multiply defined.\n" 
    }
    $$widgets{$name} = $e ;
    my $spacer = $s->Label(-font=>"ptkformVerysmall")
	    ->pack(-side=>"top", -ipady=>0, -pady=>0, -anchor=>"nw" ) ; 
    return $s ; 
}

sub ptkListbox {
    my ( $f, $ref, $label, $name, $width, $widgets) = @_ ; 
    my $s = $f->Frame() ; 
    my $l = $s->Label(-text=>$label, -font=>"ptkformLabel", -fg=>"darkblue")  
	    ->pack(-side=>"top", -ipady=>0, -pady=>0, -anchor=>"sw" ) ; 
    my $e = $s->Scrolled("Listbox", -scrollbars=>"osoe", -font=>"ptkformEntry") 
	->pack(-side=>"top", -ipady=>0, -pady=>0, -anchor=>"nw" ) ; 
    if ( $width ne "-" ) { 
	my ($rows, $columns) = split ('x', $width) ;
	$e->configure(-height=>$rows) ; 
	if ( defined $columns ) { 
	    $e->configure(-width=>$columns) ; 
	}
    }
    if ( defined $$widgets{$name} ) { 
	print STDERR "widget{$name} multiply defined.\n" 
    }
    $$widgets{$name} = $e ;
    my $spacer = $s->Label(-font=>"ptkformVerysmall")
	    ->pack(-side=>"top", -ipady=>0, -pady=>0, -anchor=>"nw" ) ; 
    return $s ; 
}

sub ptkLabel { 
    my ( $f, $ref, $label, $name, $width, $widgets) = @_ ; 
    my $s = $f->Frame() ; 
    if ( $label ne "" ) { 
	my $l = $s->Label(-text=>$label, -font=>"ptkformLabel", -fg=>"darkblue")  
		->pack(-side=>"top", -ipady=>0, -pady=>0, -anchor=>"sw" ) ; 
    }
    my $a = \$ref->$name ; 
    my $e = $s->Label(-textvariable=>$a,
	    -anchor=>"nw", -justify=>"left", -font=>"ptkformEntry")
	->pack(-side=>"top", -ipady=>0, -pady=>0, -anchor=>"nw" ) ; 
    if ( $width ne "-" ) { 
	$e->configure(-width=>$width) ; 
    }
    if ( defined $$widgets{$name} ) { 
	print STDERR "widget{$name} multiply defined.\n" 
    }
    $$widgets{$name} = $e ;
    my $spacer = $s->Label(-font=>"ptkformVerysmall")
	    ->pack(-side=>"top", -ipady=>0, -pady=>0, -anchor=>"nw" ) ; 
    return $s ; 
}

sub ptkButton { 
    my ( $f, $ref, $label, $name, $width, $widgets) = @_ ; 
    my $s = $f->Frame() ; 
    my $e = $s->Button(
	    -text=>$label, 
	    -font=>"ptkformEntry", 
	    -fg=>"darkblue" 
	    )
	->pack(-side=>"top", -ipady=>0, -pady=>0, -anchor=>"nw" ) ; 
    if ( $width ne "-" ) { 
	$e->configure(-width=>$width) ; 
    }
    if ( defined $$widgets{$name} ) { 
	print STDERR "widget{$name} multiply defined.\n" 
    }
    $$widgets{$name} = $e ;
    my $spacer = $s->Label(-font=>"ptkformVerysmall")
	    ->pack(-side=>"top", -ipady=>0, -pady=>0, -anchor=>"nw" ) ; 
    return $s ; 
}

sub ptkCheckbutton {
    my ( $f, $ref, $label, $name, $width, $widgets) = @_ ; 
    my $s = $f->Frame() ; 
    $ref->$name = 0 if ! defined $ref->$name ;
    my $a = \$ref->$name ;
    my $e = $s->Checkbutton(
	    -text=>$label, 
	    -font=>"ptkformLabel", 
	    -fg=>"darkblue",
	    -variable=>$a,
	    )
	->pack(-side=>"top", -ipady=>0, -pady=>0, -anchor=>"nw" ) ; 
    if ( $width ne "-" ) { 
	$e->configure(-width=>$width) ; 
    }
    if ( defined $$widgets{$name} ) { 
	print STDERR "widget{$name} multiply defined.\n" 
    }
    $$widgets{$name} = $e ;
    my $spacer = $s->Label(-font=>"ptkformVerysmall")
	    ->pack(-side=>"top", -ipady=>0, -pady=>0, -anchor=>"nw" ) ; 
    return $s ; 
}

sub ptkRadiobutton {
    my ( $f, $ref, $label, $name, $width, $widgets) = @_ ; 
    my $s = $f->Frame() ; 
    if ( $label =~ /^\d+\s*$/ ) { 
	$Value{$name} = $label + 0 ;
    } else {
	$Value{$name} = 0 if ! defined $Value{$name} ;
    }
    $$ref{"V$name"} = {} if ! defined $$ref{"V$name"} ;
    my $hash = $$ref{"V$name"} ; 
    $$hash{$Value{$name}} = $label ; 
    my $a = \$ref->$name ;
    $$a = 0 if ! defined $$a ;
    my $e = $s->Radiobutton(
	    -text=>$label, 
	    -font=>"ptkformLabel", 
	    -fg=>"darkblue",
	    -variable=>$a,
	    -value => $Value{$name}, 
	    )
	->pack(-side=>"top", -ipady=>0, -pady=>0, -anchor=>"nw" ) ; 
    $Value{$name} ++ ;
    if ( $width ne "-" ) { 
	$e->configure(-width=>$width) ; 
    }
    if ( ! defined $$widgets{$name} ) { 
	$$widgets{$name} = [] ;
    } 
    push ( @{$$widgets{$name}}, $e ) ; 

    my $spacer = $s->Label(-font=>"ptkformVerysmall")
	    ->pack(-side=>"top", -ipady=>0, -pady=>0, -anchor=>"nw" ) ; 
    return $s ; 
}

sub ptkOptionmenu {
    my ( $f, $ref, $label, $name, $width, $widgets) = @_ ; 
    my $s = $f->Frame() ; 
    my @list = split(' ', $label) ; 
    $label = shift(@list) ; 
    my $l = $s->Label(-text=>$label, -font=>"ptkformLabel", -fg=>"darkblue")  
	    ->pack(-side=>"top", -ipady=>0, -pady=>0, -anchor=>"sw" ) ; 
    my $a = \$ref->$name ;
    my $e = $s->Optionmenu(
	    -options=>\@list,
	    -font=>"ptkformEntry", 
	    -foreground=>"darkblue",
	    -textvariable=>$a,
	    )
	->pack(-side=>"top", -ipady=>0, -pady=>0, -anchor=>"nw" ) ; 
    if ( ! defined $ref->$name ) { 
	my $first = $list[0] ; 
	if ( $first =~ /(\w+)=>(\w+)/ ) {
	    $ref->$name = $2 ; 
	} else { 
	    $ref->$name = $first ; 
	}
    }

    if ( $width ne "-" ) { 
	$e->configure(-width=>$width) ; 
    }
    if ( defined $$widgets{$name} ) { 
	print STDERR "widget{$name} multiply defined.\n" 
    }
    $$widgets{$name} = $e ; 
    my $spacer = $s->Label(-font=>"ptkformVerysmall")
	    ->pack(-side=>"top", -ipady=>0, -pady=>0, -anchor=>"nw" ) ; 
    return $s ; 
}

sub ptkBrowseEntry {
    my ( $f, $ref, $label, $name, $width, $widgets) = @_ ; 
    my $s = $f->Frame() ; 
    my @list = split(' ', $label) ; 
    $label = shift(@list) ; 
    my $l = $s->Label(-text=>$label, -font=>"ptkformLabel", -fg=>"darkblue")  
	    ->pack(-side=>"top", -ipady=>0, -pady=>0, -anchor=>"sw" ) ; 
    my $a = \$ref->$name ;
    my $e = $s->BrowseEntry  (
	    -font=>"ptkformEntry", 
	    -variable=>$a,
	    )
	->pack(-side=>"top", -ipady=>0, -pady=>0, -anchor=>"nw" ) ; 
    if ( $width ne "-" ) { 
	$e->configure(-width=>$width) ; 
    }
    foreach my $item ( @list ) { 
	$e->insert(100, $item) ; 
    }
    if ( ! defined $ref->$name ) { 
	my $first = $list[0] ; 
	if ( $first =~ /(\w+)=>(\w+)/ ) {
	    $ref->$name = $2 ; 
	} else { 
	    $ref->$name = $first ; 
	}
    }
    if ( defined $$widgets{$name} ) { 
	print STDERR "widget{$name} multiply defined.\n" 
    }
    $$widgets{$name} = $e ; 
    my $spacer = $s->Label(-font=>"ptkformVerysmall")
	    ->pack(-side=>"top", -ipady=>0, -pady=>0, -anchor=>"nw" ) ; 
    return $s ; 
}

sub ptkFrame {
    my ( $f, $ref, $label, $name, $width, $widgets) = @_ ; 
    my $s ; 
    if ( $label ne "" ) {
	$s = $f->LabFrame(-label=>$label, -labelside=>"acrosstop") ; 
    } else { 
	$s = $f->Frame() ; 
    }
    if ( defined $$widgets{$name} ) { 
	print STDERR "widget{$name} multiply defined.\n" 
    }
    $$widgets{$name} = $s ; 
    push (@Gcol, $G{col}) ; 
    push (@Grow, $G{row}) ; 
    $G{col} = $G{row} = 0 ; 
    push (@Gframe, $f) ; 
    return $s ; 
}

sub ptkEndframe {
    my ( $f, $ref, $label, $name, $width, $widgets) = @_ ; 
    my $s ; 
    if (@Gcol > 0 ) { 
	$G{col} = pop(@Gcol) ; 
	$G{row} = pop(@Grow) ; 
	$s = pop (@Gframe) ; 
    } else { 
	print STDERR "endframe without matching frame\n" ;
	$s = $f ; 
    }
    return $s ; 
}

sub check_grid { 
    my ($grid, $name, $row, $rs, $col, $cs ) = @_ ; 
    my ($i, $j) ; 
    for ($i=0 ; $i<$rs ; $i++ ) {
	$r = $row + $i ;
	for ($j=0 ; $j<$cs ; $j++ ) {
	    $c = $col + $j ;
	    if ( defined $$grid{$r,$c} ) { 
		print STDERR "conflict: $name and $$grid{$r,$c} at  $r, $c\n" ; 
	    } else { 
		$$grid{$r,$c} = $name ; 
	    }
	}
    }
}

sub ptkform { 
    my ($f, $ref, $widgets, @spec) = @_ ;
    my $emphaticColor = "#ffff00" ; 
    my $noticeColor = "#ffffe0" ; 
    my $buttonColor = "#f0f0f0" ;
    if ( ! defined $f ) { 
	die ( "pktobjform: no window for these specs: '@spec'\n" ) ;
    }
    my $Balloon = $f->Balloon( -background=>$emphaticColor, -width=>60) ;

    mkfonts($f) ; 

    $grid = {} ; 

    my ($b, $balloon) ; 
    for(my $i=0 ; $i<@spec ; $i++) { 
	$_ = $spec[$i] ; 
	chomp() ;
	$balloon = undef ;
	next if /^\s*$/ ; 
	next if /^\s*#/ ; 
	if ( /{(.*)}/ ) { 
	    $balloon = $1 ; 
	    $_ = $` ; 
	} elsif ( /{(.*)/ ) {  # } 
	    $_ = $` ; 
	    $balloon = "$1\n" if $1 !~ /^\s*$/ ; 
	    for ( $i++ ; $i<@spec ; $i++ ) {
		$b = $spec[$i] ;  # { to match bracket below
		chomp() ;
		if ( $b =~ /^}/ ) { 
		    last ; 
		} else { 
		    $balloon .= "$b\n" ; 
		}
	    }
	    chomp($balloon) ;
	}
	my ($type, $name, $width, $rc, $label) = split(' ', $_, 5) ; 
	my ($row, $rs, $col, $cs) = parserc ( $rc ) ; 
	my $s ; 
	if ( $type eq "entry" ) { 
	    $s = ptkEntry($f, $ref, $label, $name, $width, $widgets ) ; 
	} elsif ( $type eq "frame" ) { 
	    $s = ptkFrame($f, $ref, $label, $name, $width, $widgets ) ; 
	} elsif ( $type eq "endframe" ) { 
	    $s = $f = ptkEndframe($f, $ref, $label, $name, $width, $widgets ) ; 
	    $grid = pop(@Ggrid) ;
	    # print "** endframe: grid=$grid frame=$s\n" ;
	    next ; 
	} elsif ( $type eq "optionmenu" ) { 
	    $s = ptkOptionmenu($f, $ref, $label, $name, $width, $widgets ) ; 
	} elsif ( $type eq "text" ) { 
	    $s = ptkText($f, $ref, $label, $name, $width, $widgets ) ; 
	} elsif ( $type eq "rotext" ) { 
	    $s = ptkROText($f, $ref, $label, $name, $width, $widgets ) ; 
	} elsif ( $type eq "listbox" ) { 
	    $s = ptkListbox($f, $ref, $label, $name, $width, $widgets ) ; 
	} elsif ( $type eq "button" ) { 
	    $s = ptkButton($f, $ref, $label, $name, $width, $widgets ) ; 
	} elsif ( $type eq "label" ) { 
	    $s = ptkLabel($f, $ref, $label, $name, $width, $widgets ) ; 
	} elsif ( $type eq "checkbutton" ) { 
	    $s = ptkCheckbutton($f, $ref, $label, $name, $width, $widgets ) ; 
	} elsif ( $type eq "radiobutton" ) { 
	    $s = ptkRadiobutton($f, $ref, $label, $name, $width, $widgets ) ; 
	} elsif ( $type eq "browseentry" ) { 
	    $s = ptkBrowseEntry($f, $ref, $label, $name, $width, $widgets ) ; 
	} elsif ( $type eq "hscale" ) { 
	    $s = ptkScale($f, $ref, $label, $name, $width, $widgets, "horizontal" ) ; 
	} elsif ( $type eq "vscale" ) { 
	    $s = ptkScale($f, $ref, $label, $name, $width, $widgets, "vertical" ) ; 
	} else { 
	    $s = $f->Label(-text => "unrecognized element '$type':\n'$_'" )  ;
	}
	my $depth = scalar(@Ggrid) ; 
	printf STDERR "%s(%d) %-25s -row=> %2d : %2d -col=> %2d : %2d $f\n", "  " x $depth, $depth, $name, $row, $rs, $col, $cs 
		if $Verbose ;
	check_grid($grid, $name, $row, $rs, $col, $cs ) ; 
	# if ( $type eq "frame" && $label ne "" ) { 
	#     $s->grid(-sticky=>"ewns", -row=>$row, -rowspan=>$rs, -column=>$col, -columnspan=>$cs ) ;
	# } else { 
	    $s->grid(-sticky=>$Sticky, -row=>$row, -rowspan=>$rs, -column=>$col, -columnspan=>$cs ) ;
	# }
	$Balloon->attach($s, -balloonmsg=> $balloon ) if defined $balloon ;
	if ( $type eq "frame" ) { 
	    push(@Ggrid, $grid) ;
	    $grid = {} ; 
	    $f = $s ; 
	}
    }
}

1;
